from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import NOT_PROVIDED, Model
from graphql import Undefined

from undine import Input, MutationType
from undine.converters import is_input_required
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import ModelFieldError
from undine.parsers import parse_parameters
from undine.typing import ModelField, MutationKind
from undine.utils.model_utils import get_model_field


@is_input_required.register
def _(ref: ModelField, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    is_primary_key = bool(getattr(ref, "primary_key", False))
    is_create_mutation = caller.mutation_type.__kind__ == MutationKind.create
    is_related_mutation = caller.mutation_type.__kind__ == MutationKind.related
    is_to_many_field = bool(ref.one_to_many) or bool(ref.many_to_many)
    is_nullable = bool(getattr(ref, "null", True))
    has_auto_default = bool(getattr(ref, "auto_now", False)) or bool(getattr(ref, "auto_now_add", False))
    has_default = has_auto_default or getattr(ref, "default", NOT_PROVIDED) is not NOT_PROVIDED

    if is_related_mutation:
        return False

    if is_create_mutation:
        return not is_to_many_field and not is_nullable and not has_default

    return is_primary_key


@is_input_required.register
def _(_: type[Model], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]
    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldError:
        return True
    return is_input_required(field, **kwargs)


@is_input_required.register
def _(_: TypeRef, **kwargs: Any) -> bool:
    return False


@is_input_required.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_input_required.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    parameters = parse_parameters(ref)
    first_param_default_value = next((param.default_value for param in parameters), Undefined)
    return first_param_default_value is Undefined


@is_input_required.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    return is_input_required(field, caller=caller)


@is_input_required.register
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]
    return caller.mutation_type.__kind__ == MutationKind.create



from __future__ import annotations

from types import FunctionType, GenericAlias
from typing import Any, get_origin

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import F, Model, Q, QuerySet
from graphql import GraphQLList, GraphQLNonNull, GraphQLType

from undine import Calculation, MutationType, QueryType, UnionType
from undine.converters import is_many
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.parsers import parse_return_annotation
from undine.relay import Connection
from undine.typing import CombinableExpression, ModelField
from undine.utils.model_utils import get_model_field


@is_many.register
def _(ref: ModelField, **kwargs: Any) -> bool:
    return bool(ref.many_to_many) or bool(ref.one_to_many)


@is_many.register
def _(ref: type[Model], **kwargs: Any) -> bool:
    return is_many(ref._meta.pk, **kwargs)


@is_many.register
def _(ref: TypeRef, **kwargs: Any) -> bool:
    annotation = ref.value
    if isinstance(annotation, GenericAlias):
        annotation = get_origin(annotation)
    return isinstance(annotation, type) and issubclass(annotation, list | set | tuple | QuerySet)


@is_many.register
def _(ref: CombinableExpression, **kwargs: Any) -> bool:
    return is_many(ref.output_field)


@is_many.register
def _(_: F | Q, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(ref: LazyRelation, **kwargs: Any) -> bool:
    return is_many(ref.field)


@is_many.register
def _(_: LazyGenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(ref: type[Calculation], **kwargs: Any) -> bool:
    return is_many(TypeRef(value=ref.__returns__))


@is_many.register
def _(ref: GraphQLType, **kwargs: Any) -> bool:
    return isinstance(ref, GraphQLList) or (isinstance(ref, GraphQLNonNull) and isinstance(ref.of_type, GraphQLList))


@is_many.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    annotation = parse_return_annotation(ref)
    if isinstance(annotation, GenericAlias):
        annotation = get_origin(annotation)
    return isinstance(annotation, type) and issubclass(annotation, list | set | tuple | QuerySet)


@is_many.register
def _(_: type[QueryType], **kwargs: Any) -> bool:
    field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    return is_many(field, **kwargs)


@is_many.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    return is_many(field, **kwargs)


@is_many.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(_: Connection, **kwargs: Any) -> bool:
    # Connection edges have many nodes, but the connection itself is a single node.
    return False


@is_many.register
def _(_: type[UnionType], **kwargs: Any) -> bool:
    # You always want multiple results from a union,
    # otherwise you should know its type beforehand.
    return True



from __future__ import annotations

from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey

from undine import QueryType
from undine.converters import convert_to_field_complexity
from undine.dataclasses import LazyGenericForeignKey, LazyRelation
from undine.typing import ModelField, ToManyField, ToOneField


@convert_to_field_complexity.register
def _(_: Any, **kwargs: Any) -> int:
    return 0


@convert_to_field_complexity.register
def _(_: ModelField, **kwargs: Any) -> Any:
    return 0


@convert_to_field_complexity.register
def _(_: ToOneField | ToManyField, **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register
def _(_: type[QueryType], **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register
def _(ref: LazyRelation, **kwargs: Any) -> Any:
    return convert_to_field_complexity(ref.field, **kwargs)


@convert_to_field_complexity.register
def _(ref: LazyGenericForeignKey, **kwargs: Any) -> Any:
    return convert_to_field_complexity(ref.field, **kwargs)



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, TypeGuard, TypeVar

from django.db.models import ForeignKey
from graphql import (
    DocumentNode,
    FieldNode,
    GraphQLEnumType,
    GraphQLError,
    GraphQLIncludeDirective,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSkipDirective,
    GraphQLUnionType,
    OperationDefinitionNode,
    OperationType,
    get_argument_values,
    get_directive_values,
)

from undine.exceptions import (
    DirectiveLocationError,
    GraphQLGetRequestMultipleOperationsNoOperationNameError,
    GraphQLGetRequestNonQueryOperationError,
    GraphQLGetRequestNoOperationError,
    GraphQLGetRequestOperationNotFoundError,
)
from undine.utils.text import to_snake_case

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from graphql import (
        DirectiveLocation,
        DocumentNode,
        GraphQLList,
        GraphQLNonNull,
        GraphQLOutputType,
        GraphQLWrappingType,
        SelectionNode,
    )
    from graphql.execution.values import NodeWithDirective

    from undine import Field, GQLInfo
    from undine.directives import Directive
    from undine.typing import ModelField


__all__ = [
    "check_directives",
    "get_arguments",
    "get_queried_field_name",
    "get_underlying_type",
    "is_connection",
    "is_edge",
    "is_node_interface",
    "is_page_info",
    "is_relation_id",
    "is_subscription_operation",
    "should_skip_node",
    "with_graphql_error_path",
]


TGraphQLType = TypeVar(
    "TGraphQLType",
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLInputObjectType,
)


# Getters


def get_underlying_type(
    gql_type: (
        TGraphQLType
        | GraphQLList[TGraphQLType]
        | GraphQLList[GraphQLNonNull[TGraphQLType]]
        | GraphQLNonNull[TGraphQLType]
        | GraphQLNonNull[GraphQLList[TGraphQLType]]
        | GraphQLNonNull[GraphQLList[GraphQLNonNull[TGraphQLType]]]
        | GraphQLWrappingType[TGraphQLType]
    ),
) -> TGraphQLType:
    while hasattr(gql_type, "of_type"):
        gql_type = gql_type.of_type
    return gql_type


def get_arguments(info: GQLInfo) -> dict[str, Any]:
    """Get input arguments for the current field from the GraphQL resolve info."""
    graphql_field = info.parent_type.fields[info.field_name]
    return get_argument_values(graphql_field, info.field_nodes[0], info.variable_values)


def get_queried_field_name(original_name: str, info: GQLInfo) -> str:
    """Get the name of a field in the current query."""
    return original_name if info.path.key == info.field_name else info.path.key  # type: ignore[return-value]


async def pre_evaluate_request_user(info: GQLInfo) -> None:
    """
    Fetches the request user from the context and caches it to the request.
    This is a workaround when current user is required in an async event loop,
    but the function itself is not async.
    """
    # '_current_user' would be set by 'django.contrib.auth.middleware.get_user' when calling 'request.user'
    info.context._cached_user = await info.context.auser()  # type: ignore[attr-defined]  # noqa: SLF001


# Predicates


def is_connection(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name.endswith("Connection")
        and "pageInfo" in field_type.fields
        and "edges" in field_type.fields
    )


def is_edge(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name.endswith("Edge")
        and "cursor" in field_type.fields
        and "node" in field_type.fields
    )


def is_node_interface(field_type: GraphQLOutputType) -> TypeGuard[GraphQLInterfaceType]:
    return (
        isinstance(field_type, GraphQLInterfaceType)  # comment here for better formatting
        and field_type.name == "Node"
        and "id" in field_type.fields
    )


def is_page_info(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name == "PageInfo"
        and "hasNextPage" in field_type.fields
        and "hasPreviousPage" in field_type.fields
        and "startCursor" in field_type.fields
        and "endCursor" in field_type.fields
    )


def is_typename_metafield(field_node: SelectionNode) -> TypeGuard[FieldNode]:
    if not isinstance(field_node, FieldNode):
        return False
    return field_node.name.value.lower() == "__typename"


def is_relation_id(field: ModelField, field_node: FieldNode) -> TypeGuard[Field]:
    return isinstance(field, ForeignKey) and field.get_attname() == to_snake_case(field_node.name.value)


def is_subscription_operation(document: DocumentNode) -> bool:
    if len(document.definitions) != 1:
        return False

    operation_definition = document.definitions[0]
    if not isinstance(operation_definition, OperationDefinitionNode):
        return False

    return operation_definition.operation == OperationType.SUBSCRIPTION


def should_skip_node(node: NodeWithDirective, variable_values: dict[str, Any]) -> bool:
    skip_args = get_directive_values(GraphQLSkipDirective, node, variable_values)
    if skip_args is not None and skip_args["if"] is True:
        return True

    include_args = get_directive_values(GraphQLIncludeDirective, node, variable_values)
    return include_args is not None and include_args["if"] is False


# Misc.


@contextmanager
def with_graphql_error_path(info: GQLInfo, *, key: str | int | None = None) -> Generator[None, None, None]:
    """Context manager that sets the path of all GraphQL errors raised during its context."""
    try:
        yield
    except GraphQLError as error:
        if error.path is None:
            if key is not None:
                error.path = info.path.add_key(key).as_list()
            else:
                error.path = info.path.as_list()
        raise


def check_directives(directives: Iterable[Directive] | None, *, location: DirectiveLocation) -> None:
    """Check that given directives are allowed in the given location."""
    if directives is None:
        return

    for directive in directives:
        if location not in directive.__locations__:
            raise DirectiveLocationError(directive=directive, location=location)


def validate_get_request_operation(document: DocumentNode, operation_name: str | None = None) -> None:
    """Validates that the operation in the document can be executed in an HTTP GET request."""
    operation_definitions: list[OperationDefinitionNode] = [
        definition_node
        for definition_node in document.definitions
        if isinstance(definition_node, OperationDefinitionNode)
    ]

    if len(operation_definitions) == 0:
        raise GraphQLGetRequestNoOperationError

    if operation_name is None:
        if len(operation_definitions) != 1:
            raise GraphQLGetRequestMultipleOperationsNoOperationNameError

        if operation_definitions[0].operation != OperationType.QUERY:
            raise GraphQLGetRequestNonQueryOperationError

        return

    for operation in operation_definitions:
        if operation.name is None:
            continue

        if operation.name.value != operation_name:
            continue

        if operation.operation != OperationType.QUERY:
            raise GraphQLGetRequestNonQueryOperationError

        return

    raise GraphQLGetRequestOperationNotFoundError(operation_name=operation_name)
